@@ -120,7 +120,7 @@ static const ev_map_t ev_map[] = {
 
 #define KERN_TASK_INFO(str)	{ str, sizeof(str) - 1 }
 
-static bool stop_recv;				/* sighandler stop flag */
+static volatile bool stop_recv;			/* sighandler stop flag */
 static bool sane_procs;				/* true if not inside a container */
 static proc_info_t *proc_info[MAX_PIDS];	/* Proc hash table */
 static proc_stats_t *proc_stats[MAX_PIDS];	/* Proc stats hash table */
@@ -137,6 +137,71 @@ static proc_info_t no_info = {
 	.next = NULL,
 };
 
+/*
+ *  Attempt to catch a range of signals so
+ *  we can clean
+ */
+static const int signals[] = {
+	/* POSIX.1-1990 */
+#ifdef SIGHUP
+	SIGHUP,
+#endif
+#ifdef SIGINT
+	SIGINT,
+#endif
+#ifdef SIGQUIT
+	SIGQUIT,
+#endif
+#ifdef SIGILL
+	SIGILL,
+#endif
+#ifdef SIGABRT
+	SIGABRT,
+#endif
+#ifdef SIGFPE
+	SIGFPE,
+#endif
+#ifdef SIGSEGV
+	SIGSEGV,
+#endif
+#ifdef SIGTERM
+	SIGTERM,
+#endif
+#ifdef SIGUSR1
+	SIGUSR1,
+#endif
+#ifdef SIGUSR2
+	SIGUSR2,
+	/* POSIX.1-2001 */
+#endif
+#ifdef SIGBUS
+	SIGBUS,
+#endif
+#ifdef SIGXCPU
+	SIGXCPU,
+#endif
+#ifdef SIGXFSZ
+	SIGXFSZ,
+#endif
+	/* Linux various */
+#ifdef SIGIOT
+	SIGIOT,
+#endif
+#ifdef SIGSTKFLT
+	SIGSTKFLT,
+#endif
+#ifdef SIGPWR
+	SIGPWR,
+#endif
+#ifdef SIGINFO
+	SIGINFO,
+#endif
+#ifdef SIGVTALRM
+	SIGVTALRM,
+#endif
+	-1,
+};
+
 static proc_info_t *proc_info_get(pid_t pid);
 
 /*
@@ -175,7 +240,7 @@ static bool sane_proc_pid_info(void)
 		}
 	}
 
-	fclose(fp);
+	(void)fclose(fp);
 
 	return ret;
 }
@@ -347,6 +412,7 @@ static void proc_stats_account(const pid_t pid, const event_t event)
 			stats->total++;
 			return;
 		}
+		stats = stats->next;
 	}
 	stats = calloc(1, sizeof(*stats));
 	if (stats == NULL)
@@ -372,7 +438,10 @@ int stats_cmp(const void *v1, const void *v2)
 	proc_stats_t **s1 = (proc_stats_t **)v1;
 	proc_stats_t **s2 = (proc_stats_t **)v2;
 
-	return (*s2)->total - (*s1)->total;
+	if ((*s2)->total == (*s1)->total)
+		return 0;
+
+	return ((*s2)->total > (*s1)->total) ? 1 : -1;
 }
 
 /*
@@ -461,10 +530,10 @@ static char *proc_comm(const pid_t pid)
 		return NULL;
 	}
 	if ((ret = read(fd, buffer, sizeof(buffer) - 1)) <= 0) {
-		close(fd);
+		(void)close(fd);
 		return NULL;
 	}
-	close(fd);
+	(void)close(fd);
 	buffer[ret - 1] = '\0';		/* remove trailing '\n' */
 	return strdup(buffer);
 }
@@ -487,10 +556,10 @@ static char *proc_cmdline(const pid_t pid)
 
 	memset(buffer, 0, sizeof(buffer));
 	if ((ret = read(fd, buffer, sizeof(buffer) - 1)) <= 0) {
-		close(fd);
+		(void)close(fd);
 		return proc_comm(pid);
 	}
-	close(fd);
+	(void)close(fd);
 	buffer[ret] = '\0';	/* Keeps coverity scan happy */
 
 	/*
@@ -704,10 +773,10 @@ static int proc_info_load(void)
 }
 
 /*
- *  handle_sigistop()
+ *  handle_sig()
  *	catch signal and flag a stop
  */
-static void handle_sigstop(int dummy)
+static void handle_sig(int dummy)
 {
 	(void)dummy;
 	stop_recv = true;
@@ -725,7 +794,8 @@ static int netlink_connect(void)
 	if ((sock = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_CONNECTOR)) < 0) {
 		if (errno == EPROTONOSUPPORT)
 			return -EPROTONOSUPPORT;
-		fprintf(stderr, "Socket failed: %s\n", strerror(errno));
+		fprintf(stderr, "socket failed: errno=%d (%s)\n",
+			errno, strerror(errno));
 		return -1;
 	}
 
@@ -735,7 +805,8 @@ static int netlink_connect(void)
 	addr.nl_groups = CN_IDX_PROC;
 
 	if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
-		fprintf(stderr, "Bind failed: %s\n", strerror(errno));
+		fprintf(stderr, "bind failed: errno=%d (%s)\n",
+			errno, strerror(errno));
 		(void)close(sock);
 		return -1;
 	}
@@ -777,7 +848,7 @@ static int netlink_listen(const int sock)
 
 /*
  *   monitor()
- *	monitor system activity and power consumption
+ *	monitor system activity
  */
 static int monitor(const int sock)
 {
@@ -793,22 +864,31 @@ static int monitor(const int sock)
 			return 0;
 		}
 		if (len == -1) {
-			switch (errno) {
+			int err = errno;
+			switch (err) {
 			case EINTR:
 				return 0;
 			case ENOBUFS: {
 				time_t now;
 				struct tm tm;
 
-				(void)localtime_r(&now, &tm);
-				printf("%2.2d:%2.2d:%2.2d recv ----- "
-					"nobufs %8.8s (%s)\n",
-					tm.tm_hour, tm.tm_min, tm.tm_sec, "",
-					strerror(errno));
+				now = time(NULL);
+				if (now == ((time_t) -1)) {
+					printf("--:--:-- recv ----- "
+						"nobufs %8.8s (%s)\n",
+						"", strerror(err));
+				} else {
+					(void)localtime_r(&now, &tm);
+					printf("%2.2d:%2.2d:%2.2d recv ----- "
+						"nobufs %8.8s (%s)\n",
+						tm.tm_hour, tm.tm_min, tm.tm_sec, "",
+						strerror(err));
+				}
 				break;
 			}
 			default:
-				fprintf(stderr,"recv: %d %s\n", errno, strerror(errno));
+				fprintf(stderr,"recv failed: errno=%d (%s)\n",
+					err, strerror(err));
 				return -1;
 			}
 		}
@@ -827,6 +907,9 @@ static int monitor(const int sock)
 			char *comm;
 			proc_info_t const *info1, *info2;
 
+			if (stop_recv)
+				break;
+
 			if ((nlmsghdr->nlmsg_type == NLMSG_ERROR) ||
 			    (nlmsghdr->nlmsg_type == NLMSG_NOOP))
 				continue;
@@ -838,16 +921,22 @@ static int monitor(const int sock)
 
 			proc_ev = (struct proc_event *)cn_msg->data;
 
-			(void)time(&now);
-			(void)localtime_r(&now, &tm);
-			snprintf(when, sizeof(when), "%2.2d:%2.2d:%2.2d",
-				tm.tm_hour, tm.tm_min, tm.tm_sec);
+			now = time(NULL);
+			if (now == ((time_t) -1)) {
+				snprintf(when, sizeof(when), "--:--:--");
+			} else {
+				(void)localtime_r(&now, &tm);
+				snprintf(when, sizeof(when), "%2.2d:%2.2d:%2.2d",
+					tm.tm_hour, tm.tm_min, tm.tm_sec);
+			}
 
 			switch (proc_ev->what) {
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
 			case PROC_EVENT_FORK:
 				proc_stats_account(proc_ev->event_data.fork.parent_pid, STAT_FORK);
-				gettimeofday(&tv, NULL);
+				if (gettimeofday(&tv, NULL) < 0) {
+					memset(&tv, 0, sizeof tv);
+				}
 				info1 = proc_info_get(proc_ev->event_data.fork.parent_pid);
 				info2 = proc_info_add(proc_ev->event_data.fork.child_pid, &tv);
 				if (!(opt_flags & OPT_QUIET) && (opt_flags & OPT_EV_FORK)) {
@@ -892,7 +981,9 @@ static int monitor(const int sock)
 					if (info1->start.tv_sec) {
 						double d1, d2;
 
-						gettimeofday(&tv, NULL);
+						if (gettimeofday(&tv, NULL) < 0) {
+							memset(&tv, 0, sizeof tv);
+						}
 						d1 = timeval_to_double(&info1->start);
 						d2 = timeval_to_double(&tv);
 						snprintf(duration, sizeof(duration), "%8.3f", d2 - d1);
@@ -999,11 +1090,8 @@ static int parse_ev(const char *arg)
 
 int main(int argc, char * const argv[])
 {
-	int sock = -1, ret = EXIT_FAILURE;
-
-	signal(SIGINT, &handle_sigstop);
-	signal(SIGALRM, &handle_sigstop);
-	siginterrupt(SIGINT, 1);
+	int i, sock = -1, ret = EXIT_FAILURE;
+	struct sigaction new_action;
 
 	for (;;) {
 		int c = getopt(argc, argv, "dD:e:hsSq");
@@ -1051,6 +1139,19 @@ int main(int argc, char * const argv[])
 		goto abort_sock;
 	}
 
+	memset(&new_action, 0, sizeof(new_action));
+	for (i = 0; signals[i] != -1; i++) {
+		new_action.sa_handler = handle_sig;
+		sigemptyset(&new_action.sa_mask);
+		new_action.sa_flags = 0;
+
+		if (sigaction(signals[i], &new_action, NULL) < 0) {
+			fprintf(stderr, "sigaction failed: errno=%d (%s)\n",
+				errno, strerror(errno));
+			goto abort_sock;
+		}
+	}
+
 	sane_procs = sane_proc_pid_info();
 
 	if (proc_info_load() < 0) {
@@ -1068,7 +1169,8 @@ int main(int argc, char * const argv[])
 		goto abort_sock;
 
 	if (netlink_listen(sock) < 0) {
-		fprintf(stderr, "Netlink listen failed: %s\n", strerror(errno));
+		fprintf(stderr, "netlink listen failed: errno=%d (%s)\n",
+			errno, strerror(errno));
 		goto close_abort;
 	}
 
@@ -1081,7 +1183,7 @@ int main(int argc, char * const argv[])
 	}
 
 close_abort:
-	close(sock);
+	(void)close(sock);
 abort_sock:
 	proc_info_unload();
 	proc_stats_free();
